from copy import deepcopy

import rastervision as rv
from rastervision.data.raster_source.rasterized_source import (
    RasterizedSource)
from rastervision.data.raster_source.raster_source_config \
    import (RasterSourceConfig, RasterSourceConfigBuilder)
from rastervision.protos.raster_source_pb2 \
    import RasterSourceConfig as RasterSourceConfigMsg
from rastervision.data.vector_source import VectorSourceConfig
from rastervision.utils.files import download_if_needed


class RasterizedSourceConfig(RasterSourceConfig):
    class RasterizerOptions(object):
        def __init__(self, background_class_id, all_touched=False):
            """Constructor.

            Args:
                background_class_id: The class_id to use for background pixels that don't
                    overlap with any shapes in the GeoJSON file.
                all_touched: If True, all pixels touched by geometries will be burned in.
                             If false, only pixels whose center is within the polygon or
                             that are selected by Bresenham’s line algorithm will be
                             burned in. (See rasterio.features.rasterize).
            """
            self.background_class_id = background_class_id
            self.all_touched = all_touched

        def to_proto(self):
            return RasterSourceConfigMsg.RasterizedSource.RasterizerOptions(
                background_class_id=self.background_class_id,
                all_touched=self.all_touched)

    def __init__(self,
                 vector_source,
                 rasterizer_options,
                 transformers=None,
                 channel_order=None):
        super().__init__(
            source_type=rv.RASTERIZED_SOURCE,
            transformers=transformers,
            channel_order=channel_order)
        self.vector_source = vector_source
        self.rasterizer_options = rasterizer_options

    def to_proto(self):
        msg = super().to_proto()
        msg.MergeFrom(
            RasterSourceConfigMsg(
                rasterized_source=RasterSourceConfigMsg.RasterizedSource(
                    vector_source=self.vector_source.to_proto(),
                    rasterizer_options=self.rasterizer_options.to_proto())))
        return msg

    def save_bundle_files(self, bundle_dir):
        (conf, files) = super().save_bundle_files(bundle_dir)
        # Replace the URI with a template value.
        new_config = conf.to_builder() \
                         .with_uri('BUNDLE') \
                         .build()
        return (new_config, files)

    def for_prediction(self, uri):
        return self.to_builder() \
                   .with_uri(uri) \
                   .build()

    def create_local(self, tmp_dir):
        new_uri = download_if_needed(self.uri, tmp_dir)
        return self.to_builder() \
                   .with_uri(new_uri) \
                   .build()

    def create_source(self, tmp_dir, crs_transformer, extent, class_map=None):
        vector_source = self.vector_source.create_source(
            crs_transformer=crs_transformer,
            extent=extent,
            class_map=class_map)
        return RasterizedSource(vector_source, self.rasterizer_options, extent,
                                crs_transformer)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        super().update_for_command(command_type, experiment_config, context)
        self.vector_source.update_for_command(command_type, experiment_config,
                                              context)

    def report_io(self, command_type, io_def):
        super().update_for_command(command_type, io_def)
        self.vector_source.report_io(command_type, io_def)


class RasterizedSourceConfigBuilder(RasterSourceConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'vector_source': prev.vector_source,
                'rasterizer_options': prev.rasterizer_options
            }

        super().__init__(RasterizedSourceConfig, config)

    def validate(self):
        super().validate()

        vector_source = self.config.get('vector_source')
        if vector_source is None:
            raise rv.ConfigError(
                'You must specify a vector_source for the RasterizedSourceConfig. '
                'Use "with_vector_source"')
        if not isinstance(vector_source, VectorSourceConfig):
            raise rv.ConfigError(
                'vector source must be a child of class VectorSourceConfig, got {}'.
                format(type(vector_source)))
        if vector_source.has_null_class_bufs():
            raise rv.ConfigError(
                'Setting buffer to None for a class in the vector_source is not allowed '
                'for RasterizedSourceConfig.')

        if self.config.get('rasterizer_options') is None:
            raise rv.ConfigError(
                'You must configure the rasterizer for the RasterizedSourceConfig. '
                'Use "with_rasterizer_options"')
        if not isinstance(
                self.config.get('rasterizer_options'),
                RasterizedSourceConfig.RasterizerOptions):
            raise rv.ConfigError(
                'rasterizer_options must be of type '
                'GeoJSONSourceConfig.RasterizerOptions, got'.format(
                    type(self.config.get('rasterizer_options'))))

    def from_proto(self, msg):
        b = super().from_proto(msg)

        # Added for backwards compatibility.
        if msg.HasField('geojson_file'):
            vector_source = msg.geojson_file.uri
            rasterizer_options = msg.geojson_file.rasterizer_options
        else:
            vector_source = VectorSourceConfig.from_proto(
                msg.rasterized_source.vector_source)
            rasterizer_options = msg.rasterized_source.rasterizer_options

        return b \
            .with_vector_source(vector_source) \
            .with_rasterizer_options(
                rasterizer_options.background_class_id,
                rasterizer_options.all_touched)

    def with_vector_source(self, vector_source):
        """Set the vector_source.

        Args:
            vector_source (str or VectorSource) if a string, assume it is
                a URI and use the default provider to construct a VectorSource.
        """
        if isinstance(vector_source, str):
            return self.with_uri(vector_source)

        b = deepcopy(self)
        if isinstance(vector_source, VectorSourceConfig):
            b.config['vector_source'] = vector_source
        else:
            raise rv.ConfigError(
                'vector_source must be of type str or VectorSource')

        return b

    def with_uri(self, uri):
        b = deepcopy(self)
        provider = rv._registry.get_vector_source_default_provider(uri)
        b.config['vector_source'] = provider.construct(uri)
        return b

    def with_rasterizer_options(self, background_class_id, all_touched=False):
        """Specify options for converting GeoJSON to raster.

        Args:
            background_class_id: The class_id to use for background pixels that don't
                overlap with any shapes in the GeoJSON file.
            all_touched: If True, all pixels touched by geometries will be burned in.
                         If false, only pixels whose center is within the polygon or that
                         are selected by Bresenham’s line algorithm will be burned in.
                         (See rasterio.features.rasterize).

        """
        b = deepcopy(self)
        b.config[
            'rasterizer_options'] = RasterizedSourceConfig.RasterizerOptions(
                background_class_id, all_touched)
        return b



import os
from pathlib import Path
from subprocess import Popen
import logging

from rastervision.utils.misc import terminate_at_exit
from rastervision.backend.keras_classification.utils import make_dir
log = logging.getLogger(__name__)


def get_nb_images(image_dir):
    count = 0

    pathlist = Path(image_dir).glob('**/*.png')
    for path in pathlist:
        count += 1

    pathlist = Path(image_dir).glob('**/*.jpg')
    for path in pathlist:
        count += 1

    return count


class Trainer(object):
    def __init__(self, model, optimizer, options):
        self.model = model
        self.optimizer = optimizer
        self.options = options

        self.model_path = os.path.join(options.output_dir, 'model')
        make_dir(self.model_path, use_dirname=True)
        self.weights_path = os.path.join(options.output_dir,
                                         'model-weights.hdf5')
        make_dir(self.weights_path, use_dirname=True)
        self.log_path = os.path.join(options.output_dir, 'log.csv')
        make_dir(self.log_path, use_dirname=True)

        self.training_gen = self.make_data_generator(options.training_data_dir)
        self.validation_gen = self.make_data_generator(
            options.validation_data_dir, validation_mode=True)

        self.nb_training_samples = get_nb_images(options.training_data_dir)
        self.nb_validation_samples = get_nb_images(options.validation_data_dir)

        self.tf_logs_path = os.path.join(options.output_dir, 'logs')

    def make_callbacks(self, do_monitoring):
        import keras

        model_checkpoint = keras.callbacks.ModelCheckpoint(
            filepath=self.model_path,
            save_best_only=self.options.save_best,
            save_weights_only=False)

        weights_checkpoint = keras.callbacks.ModelCheckpoint(
            filepath=self.weights_path,
            save_best_only=self.options.save_best,
            save_weights_only=True)

        csv_logger = keras.callbacks.CSVLogger(self.log_path, append=True)

        callbacks = [model_checkpoint, weights_checkpoint, csv_logger]

        if self.options.lr_schedule:
            lr_schedule = sorted(
                self.options.lr_schedule, key=lambda x: x.epoch, reverse=True)

            def schedule(curr_epoch):
                for lr_schedule_item in lr_schedule:
                    if curr_epoch >= lr_schedule_item.epoch:
                        if self.options.debug:
                            log.info('New lr: {}'.format(lr_schedule_item.lr))
                        return lr_schedule_item.lr

            lr_scheduler = keras.callbacks.LearningRateScheduler(schedule)

            callbacks.append(lr_scheduler)

        if do_monitoring:
            tensorboard = keras.callbacks.TensorBoard(
                log_dir=self.tf_logs_path, write_images=True)

            callbacks.append(tensorboard)

        return callbacks

    def get_initial_epoch(self):
        """Get initial_epoch from the last line in the log CSV file."""
        initial_epoch = 0
        if os.path.isfile(self.log_path):
            with open(self.log_path, 'r') as log_file:
                line_ind = 0
                for line_ind, _ in enumerate(log_file):
                    pass
                initial_epoch = line_ind

        return initial_epoch

    def make_data_generator(self, image_folder_dir, validation_mode=False):
        from keras.preprocessing.image import ImageDataGenerator

        # Don't apply randomized data transforms if in validation mode.
        # This will make the validation scores more comparable between epochs.
        if validation_mode:
            generator = ImageDataGenerator(rescale=1. / 255)
        else:
            generator = ImageDataGenerator(
                rescale=1. / 255, horizontal_flip=True, vertical_flip=True)

        generator = generator.flow_from_directory(
            image_folder_dir,
            classes=self.options.class_names,
            target_size=(self.options.input_size, self.options.input_size),
            batch_size=self.options.batch_size,
            class_mode='categorical')

        return generator

    def train(self, do_monitoring):
        loss_function = 'categorical_crossentropy'
        metrics = ['accuracy']
        initial_epoch = self.get_initial_epoch()
        steps_per_epoch = int(
            self.nb_training_samples / self.options.batch_size)
        validation_steps = int(
            self.nb_validation_samples / self.options.batch_size)

        # Useful for testing
        if self.options.short_epoch:
            steps_per_epoch = 1
            validation_steps = 1

        callbacks = self.make_callbacks(do_monitoring)

        self.model.compile(self.optimizer, loss_function, metrics=metrics)

        if do_monitoring:
            tensorboard_process = Popen(
                ['tensorboard', '--logdir={}'.format(self.tf_logs_path)])
            terminate_at_exit(tensorboard_process)

        self.model.fit_generator(
            self.training_gen,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            epochs=self.options.nb_epochs,
            validation_data=self.validation_gen,
            validation_steps=validation_steps,
            callbacks=callbacks)

        if do_monitoring:
            tensorboard_process.terminate()



from rastervision.backend.keras_classification.builders import optimizer_builder
from rastervision.backend.keras_classification.core.trainer import Trainer


def build(trainer_config, model):
    optimizer = optimizer_builder.build(trainer_config.optimizer)
    trainer = Trainer(model, optimizer, trainer_config.options)
    return trainer



from copy import deepcopy

import rastervision as rv
from rastervision.command import (PredictCommand, CommandConfig,
                                  CommandConfigBuilder, NoOpCommand)
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg
from rastervision.rv_config import RVConfig
from rastervision.command.utils import (check_backend_type, check_task_type)
from rastervision.utils.misc import split_into_groups


class PredictCommandConfig(CommandConfig):
    def __init__(self, root_uri, split_id, task, backend, scenes):
        super().__init__(rv.PREDICT, root_uri, split_id)
        self.task = task
        self.backend = backend
        self.scenes = scenes

    def utilizes_gpu(self):
        # TODO: Make this backend dependent.
        return True

    def create_command(self, tmp_dir=None):
        if len(self.scenes) == 0:
            return NoOpCommand()

        if not tmp_dir:
            _tmp_dir = RVConfig.get_tmp_dir()
            tmp_dir = _tmp_dir.name
        else:
            _tmp_dir = tmp_dir

        retval = PredictCommand(self)
        retval.set_tmp_dir(_tmp_dir)
        return retval

    def to_proto(self):
        msg = super().to_proto()

        task = self.task.to_proto()
        backend = self.backend.to_proto()
        scenes = list(map(lambda s: s.to_proto(), self.scenes))

        msg.MergeFrom(
            CommandConfigMsg(
                predict_config=CommandConfigMsg.PredictConfig(
                    task=task, backend=backend, scenes=scenes)))

        return msg

    def report_io(self):
        io_def = rv.core.CommandIODefinition()
        self.task.report_io(self.command_type, io_def)
        self.backend.report_io(self.command_type, io_def)
        for scene in self.scenes:
            scene.report_io(self.command_type, io_def)
        return io_def

    def split(self, num_parts):
        commands = []
        for i, l in enumerate(split_into_groups(self.scenes, num_parts)):
            c = self.to_builder() \
                .with_scenes(l) \
                .with_split_id(i) \
                .build()
            commands.append(c)
        return commands


class PredictCommandConfigBuilder(CommandConfigBuilder):
    def __init__(self, command_type, prev=None):
        super().__init__(command_type, prev)
        if prev is None:
            self.task = None
            self.backend = None
            self.scenes = []
        else:
            self.task = prev.task
            self.backend = prev.backend
            self.scenes = prev.scenes

    def validate(self):
        super().validate()
        if self.task is None:
            raise rv.ConfigError('Task not set for PredictCommandConfig. Use '
                                 'with_task or with_experiment')
        check_task_type(self.task)

        if self.backend is None:
            raise rv.ConfigError(
                'Backend not set for PredictCommandConfig. Use '
                'with_backend or with_experiment')
        check_backend_type(self.backend)

    def build(self):
        self.validate()
        return PredictCommandConfig(self.root_uri, self.split_id, self.task,
                                    self.backend, self.scenes)

    def from_proto(self, msg):
        b = super().from_proto(msg)

        conf = msg.predict_config

        task = rv.TaskConfig.from_proto(conf.task)
        backend = rv.BackendConfig.from_proto(conf.backend)
        scenes = list(map(rv.SceneConfig.from_proto, conf.scenes))

        b = b.with_task(task)
        b = b.with_backend(backend)
        b = b.with_scenes(scenes)

        return b

    def get_root_uri(self, experiment_config):
        return experiment_config.predict_uri

    def with_experiment(self, experiment_config):
        b = super().with_experiment(experiment_config)
        b = b.with_task(experiment_config.task)
        b = b.with_backend(experiment_config.backend)
        b = b.with_scenes(experiment_config.dataset.validation_scenes +
                          experiment_config.dataset.test_scenes)
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = deepcopy(self)
        b.task = task
        return b

    def with_backend(self, backend):
        b = deepcopy(self)
        b.backend = backend
        return b

    def with_scenes(self, scenes):
        b = deepcopy(self)
        b.scenes = scenes
        return b



# flake8: noqa

from rastervision.data.label_source.label_source import *
from rastervision.data.label_source.label_source_config import *
from rastervision.data.label_source.chip_classification_label_source import *
from rastervision.data.label_source.chip_classification_label_source_config import *
from rastervision.data.label_source.object_detection_label_source import *
from rastervision.data.label_source.object_detection_label_source_config import *
from rastervision.data.label_source.segmentation_class_transformer import SegmentationClassTransformer
from rastervision.data.label_source.semantic_segmentation_label_source import *
from rastervision.data.label_source.semantic_segmentation_label_source_config import *



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/ssd_anchor_generator.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/ssd_anchor_generator.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\nBrastervision/protos/tf_object_detection/ssd_anchor_generator.proto\x12\'rastervision.protos.tf_object_detection\"\xf2\x02\n\x12SsdAnchorGenerator\x12\x15\n\nnum_layers\x18\x01 \x01(\x05:\x01\x36\x12\x16\n\tmin_scale\x18\x02 \x01(\x02:\x03\x30.2\x12\x17\n\tmax_scale\x18\x03 \x01(\x02:\x04\x30.95\x12\x0e\n\x06scales\x18\x0c \x03(\x02\x12\x15\n\raspect_ratios\x18\x04 \x03(\x02\x12*\n\x1finterpolated_scale_aspect_ratio\x18\r \x01(\x02:\x01\x31\x12*\n\x1creduce_boxes_in_lowest_layer\x18\x05 \x01(\x08:\x04true\x12\x1d\n\x12\x62\x61se_anchor_height\x18\x06 \x01(\x02:\x01\x31\x12\x1c\n\x11\x62\x61se_anchor_width\x18\x07 \x01(\x02:\x01\x31\x12\x15\n\rheight_stride\x18\x08 \x03(\x05\x12\x14\n\x0cwidth_stride\x18\t \x03(\x05\x12\x15\n\rheight_offset\x18\n \x03(\x05\x12\x14\n\x0cwidth_offset\x18\x0b \x03(\x05')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_SSDANCHORGENERATOR = _descriptor.Descriptor(
  name='SsdAnchorGenerator',
  full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_layers', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.num_layers', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=6,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min_scale', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.min_scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.2),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_scale', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.max_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.95),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='scales', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.scales', index=3,
      number=12, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aspect_ratios', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.aspect_ratios', index=4,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='interpolated_scale_aspect_ratio', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.interpolated_scale_aspect_ratio', index=5,
      number=13, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reduce_boxes_in_lowest_layer', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.reduce_boxes_in_lowest_layer', index=6,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='base_anchor_height', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.base_anchor_height', index=7,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='base_anchor_width', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.base_anchor_width', index=8,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height_stride', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.height_stride', index=9,
      number=8, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='width_stride', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.width_stride', index=10,
      number=9, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height_offset', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.height_offset', index=11,
      number=10, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='width_offset', full_name='rastervision.protos.tf_object_detection.SsdAnchorGenerator.width_offset', index=12,
      number=11, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=112,
  serialized_end=482,
)

DESCRIPTOR.message_types_by_name['SsdAnchorGenerator'] = _SSDANCHORGENERATOR

SsdAnchorGenerator = _reflection.GeneratedProtocolMessageType('SsdAnchorGenerator', (_message.Message,), dict(
  DESCRIPTOR = _SSDANCHORGENERATOR,
  __module__ = 'rastervision.protos.tf_object_detection.ssd_anchor_generator_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.SsdAnchorGenerator)
  ))
_sym_db.RegisterMessage(SsdAnchorGenerator)


# @@protoc_insertion_point(module_scope)



from copy import deepcopy

import rastervision as rv
from rastervision.data.label_source import (
    LabelSourceConfig, LabelSourceConfigBuilder, ObjectDetectionLabelSource)
from rastervision.data.vector_source import VectorSourceConfig
from rastervision.protos.label_source_pb2 import LabelSourceConfig as LabelSourceConfigMsg


class ObjectDetectionLabelSourceConfig(LabelSourceConfig):
    def __init__(self, vector_source):
        super().__init__(source_type=rv.OBJECT_DETECTION)
        self.vector_source = vector_source

    def to_proto(self):
        msg = super().to_proto()
        opts = LabelSourceConfigMsg.ObjectDetectionLabelSource(
            vector_source=self.vector_source.to_proto())
        msg.object_detection_label_source.CopyFrom(opts)
        return msg

    def create_source(self, task_config, extent, crs_transformer, tmp_dir):
        vector_source = self.vector_source.create_source(
            crs_transformer=crs_transformer,
            extent=extent,
            class_map=task_config.class_map)
        return ObjectDetectionLabelSource(vector_source, crs_transformer,
                                          task_config.class_map, extent)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        self.vector_source.update_for_command(command_type, experiment_config,
                                              context)

    def report_io(self, command_type, io_def):
        self.vector_source.report_io(command_type, io_def)


class ObjectDetectionLabelSourceConfigBuilder(LabelSourceConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {'vector_source': prev.vector_source}

        super().__init__(ObjectDetectionLabelSourceConfig, config)

    def validate(self):
        super().validate()
        vector_source = self.config.get('vector_source')
        if vector_source is None:
            raise rv.ConfigError(
                'You must set the vector_source for ObjectDetectionLabelSourceConfig'
                ' Use "with_vector_source".')
        if not isinstance(vector_source, VectorSourceConfig):
            raise rv.ConfigError(
                'vector source must be a child of class VectorSourceConfig, got {}'.
                format(type(vector_source)))
        if vector_source.has_null_class_bufs():
            raise rv.ConfigError(
                'Setting buffer to None for a class in the vector_source is not allowed '
                'for ObjectDetectionLabelSourceConfig.')

    def from_proto(self, msg):
        b = ObjectDetectionLabelSourceConfigBuilder()

        # Added for backwards compatibility.
        if msg.HasField('object_detection_geojson_source'):
            vector_source = msg.object_detection_geojson_source.uri
        else:
            vector_source = rv.VectorSourceConfig.from_proto(
                msg.object_detection_label_source.vector_source)

        return b.with_vector_source(vector_source)

    def with_vector_source(self, vector_source):
        """Set the vector_source.

        Args:
            vector_source (str or VectorSource) if a string, assume it is
                a URI and use the default provider to construct a VectorSource.
        """
        if isinstance(vector_source, str):
            return self.with_uri(vector_source)

        b = deepcopy(self)
        if isinstance(vector_source, VectorSourceConfig):
            b.config['vector_source'] = vector_source
        else:
            raise rv.ConfigError(
                'vector_source must be of type str or VectorSource')

        return b

    def with_uri(self, uri):
        b = deepcopy(self)
        provider = rv._registry.get_vector_source_default_provider(uri)
        b.config['vector_source'] = provider.construct(uri)
        return b



from abc import (ABC, abstractmethod)


class Labels(ABC):
    """A set of spatially referenced labels.

    A set of labels predicted by a model or provided by human labelers for the
    sake of training.
    """

    @abstractmethod
    def __add__(self, other):
        """Add labels to these labels.

        Returns a concatenation of this and the other labels.
        """
        pass

    @abstractmethod
    def filter_by_aoi(self, aoi_polygons):
        """Returns a copy of these labels filtered by a given set of AOI polygons

        Args:
          aoi_polygons - A list of AOI polygons to filter by, in pixel coordinates.
        """
        pass

    @abstractmethod
    def __eq__(self, other):
        pass



from typing import List
import logging
from copy import deepcopy
from collections import defaultdict

import rastervision as rv

log = logging.getLogger(__name__)


class CommandDefinition:
    def __init__(self, experiment_id, command_config, io_def):
        self.experiment_id = experiment_id
        self.command_config = command_config
        self.io_def = io_def

    def _key(self):
        return (self.command_config.command_type, self.command_config.split_id,
                '|'.join(sorted(self.io_def.input_uris)), '|'.join(
                    sorted(self.io_def.output_uris)))

    def __eq__(self, other):
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    @classmethod
    def from_experiments(cls,
                         experiments: List[rv.ExperimentConfig],
                         commands_to_run: List[str],
                         splits: int = 1):
        command_definitions = []

        for experiment in experiments:
            e = deepcopy(experiment)
            log.debug(
                'Generating command definitions for experiment {}...'.format(
                    e.id))
            commands_to_update = rv.all_commands() + \
                                 list(set(commands_to_run) - set(rv.all_commands()))
            for command_type in commands_to_update:
                log.debug(
                    'Updating config for command {}...'.format(command_type))
                e.update_for_command(command_type, e)
                if command_type in commands_to_run:
                    log.debug(
                        'Creating command configurations for {}...'.format(
                            command_type))
                    base_command_config = e.make_command_config(command_type)
                    for command_config in base_command_config.split(splits):
                        io_def = command_config.report_io()
                        command_def = cls(e.id, command_config, io_def)
                        command_definitions.append(command_def)

        return command_definitions

    @classmethod
    def from_command_configs(cls,
                             command_configs: List[rv.CommandConfig],
                             commands_to_run: List[str],
                             splits: int = 1):
        command_definitions = []

        for base_command_config in command_configs:
            if base_command_config.command_type in commands_to_run:
                log.debug('Creating command configurations for {}...'.format(
                    base_command_config.command_type))
                for command_config in base_command_config.split(splits):
                    io_def = command_config.report_io()
                    command_def = cls(None, command_config, io_def)
                    command_definitions.append(command_def)

        return command_definitions

    @staticmethod
    def filter_no_output(command_definitions):
        """Filters commands that have no output."""
        result = []
        skipped = []
        for command_def in command_definitions:
            if any(command_def.io_def.output_uris):
                result.append(command_def)
            else:
                skipped.append(command_def)

        return (result, skipped)

    @staticmethod
    def remove_duplicates(command_definitions):
        """Remove duplicate commands.

        Removes duplicated commands, defining equality for a command by
        the tuple (command_type, input_uris, output_uris)
        """

        unique_commands = []
        skipped_commands = []
        seen_commands = set([])
        for command_def in command_definitions:
            if command_def not in seen_commands:
                seen_commands.add(command_def)
                unique_commands.append(command_def)
            else:
                skipped_commands.append(command_def)

        return (unique_commands, skipped_commands)

    @staticmethod
    def get_missing_inputs(command_definitions):
        """Gathers missing inputs from a set of commands.

        Returns a dictionary of experiment id to list of missing input URIs.
        """
        missing_inputs = {}
        for command_def in command_definitions:
            if command_def.io_def.missing_input_messages:
                mi = command_def.io_def.missing_input_messages
                missing_inputs[command_def.experiment_id] = mi
        return missing_inputs

    @staticmethod
    def get_clashing_commands(command_definitions):
        """Reports commands that will overwrite each other's outputs.

        Only reports commands as clashing if they are of the same command type.

        Returns a List[str, List[CommandDefinition]] of output URIs
        and clashing commands.
        """
        outputs_to_defs = defaultdict(lambda: [])
        clashing_commands = []
        for command_def in command_definitions:
            command_type = command_def.command_config.command_type
            split_id = command_def.command_config.split_id
            for output_uri in command_def.io_def.output_uris:
                outputs_to_defs[(output_uri, command_type,
                                 split_id)].append(command_def)

        for ((output_uri, _, _), command_defs) in outputs_to_defs.items():
            if len(command_defs) > 1:
                clashing_commands.append((output_uri, command_defs))

        return clashing_commands



# flake8: noqa
import logging

from rastervision.core import ConfigError
from rastervision.analyzer.api import *
from rastervision.augmentor.api import *
from rastervision.backend.api import *
from rastervision.command.api import *
from rastervision.data.api import *
from rastervision.evaluation.api import *
from rastervision.experiment.api import *
from rastervision.runner.api import *
from rastervision.task.api import *

from rastervision.predictor import Predictor

from rastervision.cli.main import main

from rastervision.registry import Registry

root_logger = logging.getLogger('rastervision')
sh = logging.StreamHandler()
sh.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s:%(name)s: %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S')
sh.setFormatter(formatter)
root_logger.addHandler(sh)

_registry = None


def _initialize():
    global _registry

    if not _registry:
        _registry = Registry()


_initialize()



from torchvision import models
from torch import nn


def get_model(model_arch, num_labels, pretrained=True):
    model = getattr(models, model_arch)(pretrained=True, progress=True)
    model.fc = nn.Linear(model.fc.in_features, num_labels)
    return model



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/optimizer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/optimizer.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n7rastervision/protos/tf_object_detection/optimizer.proto\x12\'rastervision.protos.tf_object_detection\"\xe5\x02\n\tOptimizer\x12W\n\x12rms_prop_optimizer\x18\x01 \x01(\x0b\x32\x39.rastervision.protos.tf_object_detection.RMSPropOptimizerH\x00\x12X\n\x12momentum_optimizer\x18\x02 \x01(\x0b\x32:.rastervision.protos.tf_object_detection.MomentumOptimizerH\x00\x12P\n\x0e\x61\x64\x61m_optimizer\x18\x03 \x01(\x0b\x32\x36.rastervision.protos.tf_object_detection.AdamOptimizerH\x00\x12 \n\x12use_moving_average\x18\x04 \x01(\x08:\x04true\x12$\n\x14moving_average_decay\x18\x05 \x01(\x02:\x06\x30.9999B\x0b\n\toptimizer\"\xaf\x01\n\x10RMSPropOptimizer\x12L\n\rlearning_rate\x18\x01 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.LearningRate\x12%\n\x18momentum_optimizer_value\x18\x02 \x01(\x02:\x03\x30.9\x12\x12\n\x05\x64\x65\x63\x61y\x18\x03 \x01(\x02:\x03\x30.9\x12\x12\n\x07\x65psilon\x18\x04 \x01(\x02:\x01\x31\"\x88\x01\n\x11MomentumOptimizer\x12L\n\rlearning_rate\x18\x01 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.LearningRate\x12%\n\x18momentum_optimizer_value\x18\x02 \x01(\x02:\x03\x30.9\"]\n\rAdamOptimizer\x12L\n\rlearning_rate\x18\x01 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.LearningRate\"\xc0\x03\n\x0cLearningRate\x12_\n\x16\x63onstant_learning_rate\x18\x01 \x01(\x0b\x32=.rastervision.protos.tf_object_detection.ConstantLearningRateH\x00\x12p\n\x1f\x65xponential_decay_learning_rate\x18\x02 \x01(\x0b\x32\x45.rastervision.protos.tf_object_detection.ExponentialDecayLearningRateH\x00\x12\x64\n\x19manual_step_learning_rate\x18\x03 \x01(\x0b\x32?.rastervision.protos.tf_object_detection.ManualStepLearningRateH\x00\x12\x66\n\x1a\x63osine_decay_learning_rate\x18\x04 \x01(\x0b\x32@.rastervision.protos.tf_object_detection.CosineDecayLearningRateH\x00\x42\x0f\n\rlearning_rate\"4\n\x14\x43onstantLearningRate\x12\x1c\n\rlearning_rate\x18\x01 \x01(\x02:\x05\x30.002\"\xef\x01\n\x1c\x45xponentialDecayLearningRate\x12$\n\x15initial_learning_rate\x18\x01 \x01(\x02:\x05\x30.002\x12\x1c\n\x0b\x64\x65\x63\x61y_steps\x18\x02 \x01(\r:\x07\x34\x30\x30\x30\x30\x30\x30\x12\x1a\n\x0c\x64\x65\x63\x61y_factor\x18\x03 \x01(\x02:\x04\x30.95\x12\x17\n\tstaircase\x18\x04 \x01(\x08:\x04true\x12\x1f\n\x14\x62urnin_learning_rate\x18\x05 \x01(\x02:\x01\x30\x12\x17\n\x0c\x62urnin_steps\x18\x06 \x01(\r:\x01\x30\x12\x1c\n\x11min_learning_rate\x18\x07 \x01(\x02:\x01\x30\"\x81\x02\n\x16ManualStepLearningRate\x12$\n\x15initial_learning_rate\x18\x01 \x01(\x02:\x05\x30.002\x12\x66\n\x08schedule\x18\x02 \x03(\x0b\x32T.rastervision.protos.tf_object_detection.ManualStepLearningRate.LearningRateSchedule\x12\x15\n\x06warmup\x18\x03 \x01(\x08:\x05\x66\x61lse\x1a\x42\n\x14LearningRateSchedule\x12\x0c\n\x04step\x18\x01 \x01(\r\x12\x1c\n\rlearning_rate\x18\x02 \x01(\x02:\x05\x30.002\"\xbe\x01\n\x17\x43osineDecayLearningRate\x12!\n\x12learning_rate_base\x18\x01 \x01(\x02:\x05\x30.002\x12\x1c\n\x0btotal_steps\x18\x02 \x01(\r:\x07\x34\x30\x30\x30\x30\x30\x30\x12$\n\x14warmup_learning_rate\x18\x03 \x01(\x02:\x06\x30.0002\x12\x1b\n\x0cwarmup_steps\x18\x04 \x01(\r:\x05\x31\x30\x30\x30\x30\x12\x1f\n\x14hold_base_rate_steps\x18\x05 \x01(\r:\x01\x30')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_OPTIMIZER = _descriptor.Descriptor(
  name='Optimizer',
  full_name='rastervision.protos.tf_object_detection.Optimizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rms_prop_optimizer', full_name='rastervision.protos.tf_object_detection.Optimizer.rms_prop_optimizer', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='momentum_optimizer', full_name='rastervision.protos.tf_object_detection.Optimizer.momentum_optimizer', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='adam_optimizer', full_name='rastervision.protos.tf_object_detection.Optimizer.adam_optimizer', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_moving_average', full_name='rastervision.protos.tf_object_detection.Optimizer.use_moving_average', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='moving_average_decay', full_name='rastervision.protos.tf_object_detection.Optimizer.moving_average_decay', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.9999),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='optimizer', full_name='rastervision.protos.tf_object_detection.Optimizer.optimizer',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=101,
  serialized_end=458,
)


_RMSPROPOPTIMIZER = _descriptor.Descriptor(
  name='RMSPropOptimizer',
  full_name='rastervision.protos.tf_object_detection.RMSPropOptimizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.RMSPropOptimizer.learning_rate', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='momentum_optimizer_value', full_name='rastervision.protos.tf_object_detection.RMSPropOptimizer.momentum_optimizer_value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.9),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='decay', full_name='rastervision.protos.tf_object_detection.RMSPropOptimizer.decay', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.9),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='epsilon', full_name='rastervision.protos.tf_object_detection.RMSPropOptimizer.epsilon', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=461,
  serialized_end=636,
)


_MOMENTUMOPTIMIZER = _descriptor.Descriptor(
  name='MomentumOptimizer',
  full_name='rastervision.protos.tf_object_detection.MomentumOptimizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.MomentumOptimizer.learning_rate', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='momentum_optimizer_value', full_name='rastervision.protos.tf_object_detection.MomentumOptimizer.momentum_optimizer_value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.9),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=639,
  serialized_end=775,
)


_ADAMOPTIMIZER = _descriptor.Descriptor(
  name='AdamOptimizer',
  full_name='rastervision.protos.tf_object_detection.AdamOptimizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.AdamOptimizer.learning_rate', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=777,
  serialized_end=870,
)


_LEARNINGRATE = _descriptor.Descriptor(
  name='LearningRate',
  full_name='rastervision.protos.tf_object_detection.LearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='constant_learning_rate', full_name='rastervision.protos.tf_object_detection.LearningRate.constant_learning_rate', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exponential_decay_learning_rate', full_name='rastervision.protos.tf_object_detection.LearningRate.exponential_decay_learning_rate', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='manual_step_learning_rate', full_name='rastervision.protos.tf_object_detection.LearningRate.manual_step_learning_rate', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='cosine_decay_learning_rate', full_name='rastervision.protos.tf_object_detection.LearningRate.cosine_decay_learning_rate', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.LearningRate.learning_rate',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=873,
  serialized_end=1321,
)


_CONSTANTLEARNINGRATE = _descriptor.Descriptor(
  name='ConstantLearningRate',
  full_name='rastervision.protos.tf_object_detection.ConstantLearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.ConstantLearningRate.learning_rate', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1323,
  serialized_end=1375,
)


_EXPONENTIALDECAYLEARNINGRATE = _descriptor.Descriptor(
  name='ExponentialDecayLearningRate',
  full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='initial_learning_rate', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.initial_learning_rate', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='decay_steps', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.decay_steps', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=4000000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='decay_factor', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.decay_factor', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.95),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='staircase', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.staircase', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='burnin_learning_rate', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.burnin_learning_rate', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='burnin_steps', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.burnin_steps', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min_learning_rate', full_name='rastervision.protos.tf_object_detection.ExponentialDecayLearningRate.min_learning_rate', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1378,
  serialized_end=1617,
)


_MANUALSTEPLEARNINGRATE_LEARNINGRATESCHEDULE = _descriptor.Descriptor(
  name='LearningRateSchedule',
  full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.LearningRateSchedule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='step', full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.LearningRateSchedule.step', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='learning_rate', full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.LearningRateSchedule.learning_rate', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1811,
  serialized_end=1877,
)

_MANUALSTEPLEARNINGRATE = _descriptor.Descriptor(
  name='ManualStepLearningRate',
  full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='initial_learning_rate', full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.initial_learning_rate', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='schedule', full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.schedule', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='warmup', full_name='rastervision.protos.tf_object_detection.ManualStepLearningRate.warmup', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_MANUALSTEPLEARNINGRATE_LEARNINGRATESCHEDULE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1620,
  serialized_end=1877,
)


_COSINEDECAYLEARNINGRATE = _descriptor.Descriptor(
  name='CosineDecayLearningRate',
  full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='learning_rate_base', full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate.learning_rate_base', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_steps', full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate.total_steps', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=4000000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='warmup_learning_rate', full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate.warmup_learning_rate', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.0002),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='warmup_steps', full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate.warmup_steps', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=10000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hold_base_rate_steps', full_name='rastervision.protos.tf_object_detection.CosineDecayLearningRate.hold_base_rate_steps', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1880,
  serialized_end=2070,
)

_OPTIMIZER.fields_by_name['rms_prop_optimizer'].message_type = _RMSPROPOPTIMIZER
_OPTIMIZER.fields_by_name['momentum_optimizer'].message_type = _MOMENTUMOPTIMIZER
_OPTIMIZER.fields_by_name['adam_optimizer'].message_type = _ADAMOPTIMIZER
_OPTIMIZER.oneofs_by_name['optimizer'].fields.append(
  _OPTIMIZER.fields_by_name['rms_prop_optimizer'])
_OPTIMIZER.fields_by_name['rms_prop_optimizer'].containing_oneof = _OPTIMIZER.oneofs_by_name['optimizer']
_OPTIMIZER.oneofs_by_name['optimizer'].fields.append(
  _OPTIMIZER.fields_by_name['momentum_optimizer'])
_OPTIMIZER.fields_by_name['momentum_optimizer'].containing_oneof = _OPTIMIZER.oneofs_by_name['optimizer']
_OPTIMIZER.oneofs_by_name['optimizer'].fields.append(
  _OPTIMIZER.fields_by_name['adam_optimizer'])
_OPTIMIZER.fields_by_name['adam_optimizer'].containing_oneof = _OPTIMIZER.oneofs_by_name['optimizer']
_RMSPROPOPTIMIZER.fields_by_name['learning_rate'].message_type = _LEARNINGRATE
_MOMENTUMOPTIMIZER.fields_by_name['learning_rate'].message_type = _LEARNINGRATE
_ADAMOPTIMIZER.fields_by_name['learning_rate'].message_type = _LEARNINGRATE
_LEARNINGRATE.fields_by_name['constant_learning_rate'].message_type = _CONSTANTLEARNINGRATE
_LEARNINGRATE.fields_by_name['exponential_decay_learning_rate'].message_type = _EXPONENTIALDECAYLEARNINGRATE
_LEARNINGRATE.fields_by_name['manual_step_learning_rate'].message_type = _MANUALSTEPLEARNINGRATE
_LEARNINGRATE.fields_by_name['cosine_decay_learning_rate'].message_type = _COSINEDECAYLEARNINGRATE
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['constant_learning_rate'])
_LEARNINGRATE.fields_by_name['constant_learning_rate'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['exponential_decay_learning_rate'])
_LEARNINGRATE.fields_by_name['exponential_decay_learning_rate'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['manual_step_learning_rate'])
_LEARNINGRATE.fields_by_name['manual_step_learning_rate'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_LEARNINGRATE.oneofs_by_name['learning_rate'].fields.append(
  _LEARNINGRATE.fields_by_name['cosine_decay_learning_rate'])
_LEARNINGRATE.fields_by_name['cosine_decay_learning_rate'].containing_oneof = _LEARNINGRATE.oneofs_by_name['learning_rate']
_MANUALSTEPLEARNINGRATE_LEARNINGRATESCHEDULE.containing_type = _MANUALSTEPLEARNINGRATE
_MANUALSTEPLEARNINGRATE.fields_by_name['schedule'].message_type = _MANUALSTEPLEARNINGRATE_LEARNINGRATESCHEDULE
DESCRIPTOR.message_types_by_name['Optimizer'] = _OPTIMIZER
DESCRIPTOR.message_types_by_name['RMSPropOptimizer'] = _RMSPROPOPTIMIZER
DESCRIPTOR.message_types_by_name['MomentumOptimizer'] = _MOMENTUMOPTIMIZER
DESCRIPTOR.message_types_by_name['AdamOptimizer'] = _ADAMOPTIMIZER
DESCRIPTOR.message_types_by_name['LearningRate'] = _LEARNINGRATE
DESCRIPTOR.message_types_by_name['ConstantLearningRate'] = _CONSTANTLEARNINGRATE
DESCRIPTOR.message_types_by_name['ExponentialDecayLearningRate'] = _EXPONENTIALDECAYLEARNINGRATE
DESCRIPTOR.message_types_by_name['ManualStepLearningRate'] = _MANUALSTEPLEARNINGRATE
DESCRIPTOR.message_types_by_name['CosineDecayLearningRate'] = _COSINEDECAYLEARNINGRATE

Optimizer = _reflection.GeneratedProtocolMessageType('Optimizer', (_message.Message,), dict(
  DESCRIPTOR = _OPTIMIZER,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.Optimizer)
  ))
_sym_db.RegisterMessage(Optimizer)

RMSPropOptimizer = _reflection.GeneratedProtocolMessageType('RMSPropOptimizer', (_message.Message,), dict(
  DESCRIPTOR = _RMSPROPOPTIMIZER,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.RMSPropOptimizer)
  ))
_sym_db.RegisterMessage(RMSPropOptimizer)

MomentumOptimizer = _reflection.GeneratedProtocolMessageType('MomentumOptimizer', (_message.Message,), dict(
  DESCRIPTOR = _MOMENTUMOPTIMIZER,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.MomentumOptimizer)
  ))
_sym_db.RegisterMessage(MomentumOptimizer)

AdamOptimizer = _reflection.GeneratedProtocolMessageType('AdamOptimizer', (_message.Message,), dict(
  DESCRIPTOR = _ADAMOPTIMIZER,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.AdamOptimizer)
  ))
_sym_db.RegisterMessage(AdamOptimizer)

LearningRate = _reflection.GeneratedProtocolMessageType('LearningRate', (_message.Message,), dict(
  DESCRIPTOR = _LEARNINGRATE,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.LearningRate)
  ))
_sym_db.RegisterMessage(LearningRate)

ConstantLearningRate = _reflection.GeneratedProtocolMessageType('ConstantLearningRate', (_message.Message,), dict(
  DESCRIPTOR = _CONSTANTLEARNINGRATE,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ConstantLearningRate)
  ))
_sym_db.RegisterMessage(ConstantLearningRate)

ExponentialDecayLearningRate = _reflection.GeneratedProtocolMessageType('ExponentialDecayLearningRate', (_message.Message,), dict(
  DESCRIPTOR = _EXPONENTIALDECAYLEARNINGRATE,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ExponentialDecayLearningRate)
  ))
_sym_db.RegisterMessage(ExponentialDecayLearningRate)

ManualStepLearningRate = _reflection.GeneratedProtocolMessageType('ManualStepLearningRate', (_message.Message,), dict(

  LearningRateSchedule = _reflection.GeneratedProtocolMessageType('LearningRateSchedule', (_message.Message,), dict(
    DESCRIPTOR = _MANUALSTEPLEARNINGRATE_LEARNINGRATESCHEDULE,
    __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
    # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ManualStepLearningRate.LearningRateSchedule)
    ))
  ,
  DESCRIPTOR = _MANUALSTEPLEARNINGRATE,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ManualStepLearningRate)
  ))
_sym_db.RegisterMessage(ManualStepLearningRate)
_sym_db.RegisterMessage(ManualStepLearningRate.LearningRateSchedule)

CosineDecayLearningRate = _reflection.GeneratedProtocolMessageType('CosineDecayLearningRate', (_message.Message,), dict(
  DESCRIPTOR = _COSINEDECAYLEARNINGRATE,
  __module__ = 'rastervision.protos.tf_object_detection.optimizer_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.CosineDecayLearningRate)
  ))
_sym_db.RegisterMessage(CosineDecayLearningRate)


# @@protoc_insertion_point(module_scope)



import logging

import click

import rastervision as rv
from rastervision.augmentor import AugmentorConfig
from rastervision.data import (SceneConfig, Dataset)
from rastervision.core.config import (Config, ConfigBuilder)
from rastervision.protos.dataset_pb2 import DatasetConfig as DatasetConfigMsg
from rastervision.cli import Verbosity

log = logging.getLogger(__name__)


class DatasetConfig(Config):
    def __init__(self,
                 train_scenes=None,
                 validation_scenes=None,
                 test_scenes=None,
                 augmentors=None):
        if train_scenes is None:
            train_scenes = []
        if validation_scenes is None:
            validation_scenes = []
        if test_scenes is None:
            test_scenes = []
        if augmentors is None:
            augmentors = []

        self.train_scenes = train_scenes
        self.validation_scenes = validation_scenes
        self.test_scenes = test_scenes
        self.augmentors = augmentors

    def all_scenes(self):
        return self.train_scenes + \
            self.validation_scenes + \
            self.test_scenes

    def to_builder(self):
        return DatasetConfigBuilder(self)

    def create_dataset(self,
                       task_config,
                       tmp_dir,
                       include_train=True,
                       include_val=True,
                       include_test=True):
        train_scenes = []
        if include_train:
            train_scenes = list(
                map(lambda x: x.create_scene(task_config, tmp_dir),
                    self.train_scenes))

        val_scenes = []
        if include_val:
            val_scenes = list(
                map(lambda x: x.create_scene(task_config, tmp_dir),
                    self.validation_scenes))

        test_scenes = []
        if include_test:
            test_scenes = list(
                map(lambda x: x.create_scene(task_config, tmp_dir),
                    self.test_scenes))

        augmentors = list(map(lambda x: x.create_augmentor(), self.augmentors))

        return Dataset(
            train_scenes=train_scenes,
            validation_scenes=val_scenes,
            test_scenes=test_scenes,
            augmentors=augmentors)

    def to_proto(self):
        """Returns the protobuf configuration for this config.
        """
        train_scenes = list(map(lambda x: x.to_proto(), self.train_scenes))
        val_scenes = list(map(lambda x: x.to_proto(), self.validation_scenes))
        test_scenes = list(map(lambda x: x.to_proto(), self.test_scenes))

        augmentors = list(map(lambda x: x.to_proto(), self.augmentors))

        return DatasetConfigMsg(
            train_scenes=train_scenes,
            validation_scenes=val_scenes,
            test_scenes=test_scenes,
            augmentors=augmentors)

    def update_for_command(self,
                           command_type,
                           experiment_config,
                           context=None,
                           io_def=None):
        verbosity = Verbosity.get()

        def update_scenes(scenes_to_update, ensure_label_store):
            for scene in scenes_to_update:
                if ensure_label_store:
                    # Ensure there is a label store associated with
                    # validation and test scenes on PREDICT command.
                    if not scene.label_store:
                        scene.label_store = scene.to_builder() \
                                                 .with_task(experiment_config.task) \
                                                 .with_label_store() \
                                                 .build() \
                                                 .label_store
                scene.update_for_command(command_type, experiment_config,
                                         context)

        if command_type in [rv.ANALYZE, rv.CHIP]:
            log.debug(
                'Updating train scenes for command {}'.format(command_type))
            if verbosity >= Verbosity.VERBOSE:
                with click.progressbar(
                        self.train_scenes,
                        label='Updating train scenes') as scenes_to_update:
                    update_scenes(scenes_to_update, False)
            else:
                update_scenes(self.train_scenes, False)

        if command_type in [
                rv.ANALYZE, rv.CHIP, rv.PREDICT, rv.EVAL, rv.BUNDLE
        ]:
            log.debug('Updating validation scenes for command {}'.format(
                command_type))
            if Verbosity.get() >= Verbosity.VERBOSE:
                with click.progressbar(
                        self.validation_scenes,
                        label='Updating validation scenes...  '
                ) as scenes_to_update:
                    update_scenes(scenes_to_update, command_type == rv.PREDICT)
            else:
                update_scenes(self.validation_scenes,
                              command_type == rv.PREDICT)

        if command_type in [rv.ANALYZE, rv.PREDICT, rv.EVAL, rv.BUNDLE]:

            log.debug(
                'Updating test scenes for command {}'.format(command_type))
            if Verbosity.get() >= Verbosity.VERBOSE:
                with click.progressbar(
                        self.test_scenes,
                        label='Updating test scenes...  ') as scenes_to_update:
                    update_scenes(scenes_to_update, command_type == rv.PREDICT)
            else:
                update_scenes(self.test_scenes, command_type == rv.PREDICT)

        if command_type == rv.CHIP:
            log.debug(
                'Updating augmentors for command {}'.format(command_type))
            for augmentor in self.augmentors:
                augmentor.update_for_command(command_type, experiment_config,
                                             context)

    def report_io(self, command_type, io_def):
        if command_type in [rv.ANALYZE, rv.CHIP]:
            for scene in self.train_scenes:
                scene.report_io(command_type, io_def)

        if command_type in [
                rv.ANALYZE, rv.CHIP, rv.PREDICT, rv.EVAL, rv.BUNDLE
        ]:
            for scene in self.validation_scenes:
                scene.report_io(command_type, io_def)

        if command_type in [rv.ANALYZE, rv.PREDICT, rv.EVAL, rv.BUNDLE]:
            for scene in self.test_scenes:
                scene.report_io(command_type, io_def)

        if command_type == rv.CHIP:
            for augmentor in self.augmentors:
                augmentor.report_io(command_type, io_def)

        return io_def

    @staticmethod
    def from_proto(msg):
        """Creates a TaskConfig from the specificed protobuf message
        """
        return DatasetConfigBuilder().from_proto(msg).build()

    @staticmethod
    def builder():
        return DatasetConfigBuilder()


class DatasetConfigBuilder(ConfigBuilder):
    def __init__(self, prev=None):
        config = {
            'train_scenes': [],
            'validation_scenes': [],
            'test_scenes': [],
            'augmentors': []
        }
        if prev:
            config['train_scenes'] = prev.train_scenes
            config['validation_scenes'] = prev.validation_scenes
            config['test_scenes'] = prev.test_scenes
            config['augmentors'] = prev.augmentors
        super().__init__(DatasetConfig, config)

    def _copy(self):
        """Create a copy; avoid using deepcopy as it can have
        performance implications with many scenes
        """
        ds = DatasetConfigBuilder()
        ds.config['train_scenes'] = self.config['train_scenes']
        ds.config['validation_scenes'] = self.config['validation_scenes']
        ds.config['test_scenes'] = self.config['test_scenes']
        ds.config['augmentors'] = self.config['augmentors']
        return ds

    def from_proto(self, msg):
        train_scenes = list(
            map(lambda x: SceneConfig.from_proto(x), msg.train_scenes))
        val_scenes = list(
            map(lambda x: SceneConfig.from_proto(x), msg.validation_scenes))
        test_scenes = list(
            map(lambda x: SceneConfig.from_proto(x), msg.test_scenes))
        augmentors = list(
            map(lambda x: AugmentorConfig.from_proto(x), msg.augmentors))
        return self \
            .with_train_scenes(train_scenes) \
            .with_validation_scenes(val_scenes) \
            .with_test_scenes(test_scenes) \
            .with_augmentors(augmentors)

    def with_train_scenes(self, scenes):
        """Sets the scenes to be used for training."""
        b = self._copy()
        b.config['train_scenes'] = list(scenes)
        return b

    def with_train_scene(self, scene):
        """Sets the scene to be used for training."""
        return self.with_train_scenes([scene])

    def with_validation_scenes(self, scenes):
        """Sets the scenes to be used for validation."""
        b = self._copy()
        b.config['validation_scenes'] = list(scenes)
        return b

    def with_validation_scene(self, scene):
        """Sets the scene to be used for validation."""
        return self.with_validation_scenes([scene])

    def with_test_scenes(self, scenes):
        """Sets the scenes to be used for testing."""
        b = self._copy()
        b.config['test_scenes'] = list(scenes)
        return b

    def with_test_scene(self, scene):
        """Sets the scene to be used for testing."""
        return self.with_test_scenes([scene])

    def with_augmentors(self, augmentors):
        """Sets the data augmentors to be used."""
        b = self._copy()
        b.config['augmentors'] = list(augmentors)
        return b

    def with_augmentor(self, augmentor):
        """Sets the data augmentor to be used."""
        return self.with_augmentors([augmentor])



from collections import defaultdict

import torch

from torchvision.ops.boxes import batched_nms


def to_box_pixel(boxes, img_height, img_width):
    # convert from (ymin, xmin, ymax, xmax) in range [-1,1] to
    # range [0, h) or [0, w)
    boxes = ((boxes + 1.0) / 2.0) * torch.tensor(
        [[img_height, img_width, img_height, img_width]]).to(
            device=boxes.device, dtype=torch.float)
    return boxes


class BoxList():
    def __init__(self, boxes, **extras):
        """Constructor.

        Args:
            boxes: tensor<n, 4> with order ymin, xmin, ymax, xmax in pixels coords
            extras: dict with values that are tensors with first dimension corresponding
                to boxes first dimension
        """
        self.boxes = boxes
        self.extras = extras

    def get_field(self, name):
        if name == 'boxes':
            return self.boxes
        else:
            return self.extras.get(name)

    def _map_extras(self, func):
        new_extras = {}
        for k, v in self.extras.items():
            new_extras[k] = func(v)
        return new_extras

    def copy(self):
        return BoxList(self.boxes.copy(),
                       **self._map_extras(lambda x: x.copy()))

    def cpu(self):
        return BoxList(self.boxes.cpu(), **self._map_extras(lambda x: x.cpu()))

    def cuda(self):
        return BoxList(self.boxes.cuda(),
                       **self._map_extras(lambda x: x.cuda()))

    def to(self, device):
        return self.cpu() if device == 'cpu' else self.cuda()

    def xyxy(self):
        boxes = self.boxes[:, [1, 0, 3, 2]]
        return BoxList(boxes, **self.extras)

    def yxyx(self):
        boxes = self.boxes[:, [1, 0, 3, 2]]
        return BoxList(boxes, **self.extras)

    def __len__(self):
        return self.boxes.shape[0]

    @staticmethod
    def cat(box_lists):
        boxes = []
        extras = defaultdict(list)
        for bl in box_lists:
            boxes.append(bl.boxes)
            for k, v in bl.extras.items():
                extras[k].append(v)
        boxes = torch.cat(boxes)
        for k, v in extras.items():
            extras[k] = torch.cat(v)
        return BoxList(boxes, **extras)

    def equal(self, other):
        if len(other) != len(self):
            return False

        # Ignore order of boxes.
        extras = [(v.float().unsqueeze(1) if v.ndim == 1 else v.float())
                  for v in self.extras.values()]
        cat_arr = torch.cat([self.boxes] + extras, 1)
        self_tups = set([tuple([x.item() for x in row]) for row in cat_arr])

        extras = [(v.float().unsqueeze(1) if v.ndim == 1 else v.float())
                  for v in other.extras.values()]
        cat_arr = torch.cat([other.boxes] + extras, 1)
        other_tups = set([tuple([x.item() for x in row]) for row in cat_arr])
        return self_tups == other_tups

    def ind_filter(self, inds):
        new_extras = {}
        for k, v in self.extras.items():
            new_extras[k] = v[inds, ...]
        return BoxList(self.boxes[inds, :], **new_extras)

    def score_filter(self, score_thresh=0.25):
        scores = self.extras.get('scores')
        if scores is not None:
            return self.ind_filter(scores > score_thresh)
        else:
            raise ValueError('must have scores as key in extras')

    def clamp(self, img_height, img_width):
        boxes = torch.stack(
            [
                torch.clamp(self.boxes[:, 0], 0, img_height),
                torch.clamp(self.boxes[:, 1], 0, img_width),
                torch.clamp(self.boxes[:, 2], 0, img_height),
                torch.clamp(self.boxes[:, 3], 0, img_width)
            ],
            dim=1)
        return BoxList(boxes, **self.extras)

    def nms(self, iou_thresh=0.5):
        if len(self) == 0:
            return self

        good_inds = batched_nms(self.boxes, self.get_field('scores'),
                                self.get_field('labels'), iou_thresh)
        return self.ind_filter(good_inds)

    def scale(self, yscale, xscale):
        boxes = self.boxes * torch.tensor(
            [[yscale, xscale, yscale, xscale]], device=self.boxes.device)
        return BoxList(boxes, **self.extras)

    def pin_memory(self):
        self.boxes = self.boxes.pin_memory()
        for k, v in self.extras.items():
            self.extras[k] = v.pin_memory()
        return self



import os
from copy import deepcopy
import logging
import json

from google.protobuf import (json_format, struct_pb2)

import rastervision as rv
from rastervision.core.config import (Config, ConfigBuilder)
from rastervision.utils.files import save_json_config
from rastervision.protos.experiment_pb2 \
    import ExperimentConfig as ExperimentConfigMsg
from rastervision.backend import BackendConfig
from rastervision.task import TaskConfig
from rastervision.data import DatasetConfig

log = logging.getLogger(__name__)


class ExperimentConfig(Config):
    def __init__(self,
                 id,
                 task,
                 backend,
                 dataset,
                 root_uri,
                 analyze_uri,
                 chip_uri,
                 train_uri,
                 predict_uri,
                 eval_uri,
                 bundle_uri,
                 evaluators=None,
                 analyzers=None,
                 custom_config=None):
        if analyzers is None:
            analyzers = []

        self.id = id
        self.task = task
        self.backend = backend
        self.dataset = dataset
        self.analyzers = analyzers
        self.evaluators = evaluators
        self.root_uri = root_uri
        self.analyze_uri = analyze_uri
        self.chip_uri = chip_uri
        self.train_uri = train_uri
        self.predict_uri = predict_uri
        self.eval_uri = eval_uri
        self.bundle_uri = bundle_uri
        self.custom_config = custom_config or {}

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        log.debug('Updating task for command {}'.format(command_type))
        self.task.update_for_command(command_type, experiment_config, context)

        log.debug('Updating backend for command {}'.format(command_type))
        self.backend.update_for_command(command_type, experiment_config,
                                        context)

        log.debug('Updating dataset for command {}'.format(command_type))
        self.dataset.update_for_command(command_type, experiment_config,
                                        context)

        log.debug('Updating analyzers for command {}'.format(command_type))
        for analyzer in self.analyzers:
            analyzer.update_for_command(command_type, experiment_config,
                                        context)

        log.debug('Updating evaluators for command {}'.format(command_type))
        for evaluator in self.evaluators:
            evaluator.update_for_command(command_type, experiment_config,
                                         context)

    def report_io(self, command_type, io_def):
        log.debug('Reporting IO on task for command {}'.format(command_type))
        self.task.report_io(command_type, io_def)

        log.debug(
            'Reporting IO on backend for command {}'.format(command_type))
        self.backend.report_io(command_type, io_def)

        log.debug(
            'Reporting IO on dataset for command {}'.format(command_type))
        self.dataset.report_io(command_type, io_def)

        log.debug(
            'Reporting IO on analyzers for command {}'.format(command_type))
        for analyzer in self.analyzers:
            analyzer.report_io(command_type, io_def)

        log.debug(
            'Reporting IO on evaluators for command {}'.format(command_type))
        for evaluator in self.evaluators:
            evaluator.report_io_command(command_type, io_def)

    def make_command_config(self, command_type):
        return rv._registry.get_command_config_builder(command_type)() \
                           .with_experiment(self) \
                           .build()

    def fully_resolve(self):
        """Returns a fully resolved copy of this  experiment.

        A fully resolved experiment has all implicit paths put into place,
        and is constructed by calling update_for_command for each command.
        """
        e = deepcopy(self)
        for command_type in rv.all_commands():
            e.update_for_command(command_type, e)
        return e

    def save_config(self):
        msg = self.to_proto()
        uri = os.path.join(self.root_uri, 'experiments',
                           '{}.json'.format(self.id))
        save_json_config(msg, uri)

    def to_proto(self):
        analyzers = list(map(lambda a: a.to_proto(), self.analyzers))
        evaluators = list(map(lambda e: e.to_proto(), self.evaluators))

        msg = ExperimentConfigMsg(
            id=self.id,
            task=self.task.to_proto(),
            backend=self.backend.to_proto(),
            dataset=self.dataset.to_proto(),
            analyzers=analyzers,
            evaluators=evaluators)
        msg.root_uri = self.root_uri
        msg.analyze_uri = self.analyze_uri
        msg.chip_uri = self.chip_uri
        msg.train_uri = self.train_uri
        msg.predict_uri = self.predict_uri
        msg.eval_uri = self.eval_uri
        msg.bundle_uri = self.bundle_uri

        if self.custom_config:
            msg.MergeFrom(
                ExperimentConfigMsg(
                    custom_config=json_format.ParseDict(
                        {
                            'config': json.dumps(self.custom_config)
                        }, struct_pb2.Struct())))

        return msg

    def to_builder(self):
        return ExperimentConfigBuilder(self)

    @staticmethod
    def builder():
        return ExperimentConfigBuilder()

    @staticmethod
    def from_proto(msg):
        """Creates an ExperimentConfig from the specificed protobuf message
        """
        return ExperimentConfigBuilder().from_proto(msg).build()


class ExperimentConfigBuilder(ConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'id': prev.id,
                'task': prev.task,
                'backend': prev.backend,
                'dataset': prev.dataset,
                'analyzers': prev.analyzers,
                'evaluators': prev.evaluators,
                'root_uri': prev.root_uri,
                'analyze_uri': prev.analyze_uri,
                'chip_uri': prev.chip_uri,
                'train_uri': prev.train_uri,
                'predict_uri': prev.predict_uri,
                'eval_uri': prev.eval_uri,
                'bundle_uri': prev.bundle_uri,
                'custom_config': prev.custom_config
            }
        super().__init__(ExperimentConfig, config)
        self.analyze_key = None
        self.chip_key = None
        self.train_key = None
        self.predict_key = None
        self.eval_key = None
        self.bundle_key = None
        self.custom_config = None

    def validate(self):

        root_uri = self.config.get('root_uri')
        if not root_uri:
            raise rv.ConfigError('root_uri must be set. Use "with_root_uri"')
        if not isinstance(root_uri, str):
            raise rv.ConfigError(
                'root_uri needs to be of type str, got {}'.format(
                    type(root_uri)))

        for key in ['task', 'backend', 'dataset', 'id']:
            if self.config.get(key) is None:
                raise rv.ConfigError(
                    'Experiment %s must be set. Use "with_%s".' % (key, key))

        task = self.config.get('task')
        if not issubclass(type(task), TaskConfig):
            raise rv.ConfigError(
                'Experiment task set with "with_task" must be of class'
                ' TaskConfig, got {}'.format(type(task)))

        backend = self.config.get('backend')
        if not issubclass(type(backend), BackendConfig):
            raise rv.ConfigError(
                'Backend set with "with_backend" needs to be of type '
                'BackendConfig, got {}'.format(type(backend)))

        dataset = self.config.get('dataset')
        if not isinstance(dataset, DatasetConfig):
            raise rv.ConfigError(
                'Dataset set with "with_dataset" needs to be of type'
                'DatasetConfig, got {}'.format(type(dataset)))

        if not isinstance(self.config.get('id'), str):
            raise rv.ConfigError(
                'ID set with "with_id" needs to be of type str, got {}'.format(
                    type(self.config.get('id'))))

    def build(self):
        self.validate()
        # Build any missing paths through
        b = self

        if not self.config.get('analyze_uri'):
            if not self.analyze_key:
                self.analyze_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.ANALYZE.lower(),
                               self.analyze_key)
            b = b.with_analyze_uri(uri)
        if not self.config.get('chip_uri'):
            if not self.chip_key:
                self.chip_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.CHIP.lower(),
                               self.chip_key)
            b = b.with_chip_uri(uri)
        if not self.config.get('train_uri'):
            if not self.train_key:
                self.train_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.TRAIN.lower(),
                               self.train_key)
            b = b.with_train_uri(uri)
        if not self.config.get('predict_uri'):
            if not self.predict_key:
                self.predict_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.PREDICT.lower(),
                               self.predict_key)
            b = b.with_predict_uri(uri)
        if not self.config.get('eval_uri'):
            if not self.eval_key:
                self.eval_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.EVAL.lower(),
                               self.eval_key)
            b = b.with_eval_uri(uri)
        if not self.config.get('bundle_uri'):
            if not self.bundle_key:
                self.bundle_key = self.config['id']
            uri = os.path.join(self.config['root_uri'], rv.BUNDLE.lower(),
                               self.bundle_key)
            b = b.with_bundle_uri(uri)

        evaluators = self.config.get('evaluators')
        if not evaluators:
            task_type = self.config['task'].task_type
            e = rv._registry.get_evaluator_default_provider(task_type) \
                            .construct(self.config['task'])
            b = b.with_evaluator(e)

        return ExperimentConfig(**b.config)

    def from_proto(self, msg):
        analyzers = list(
            map(lambda a: rv.AnalyzerConfig.from_proto(a), msg.analyzers))
        evaluators = list(
            map(lambda e: rv.EvaluatorConfig.from_proto(e), msg.evaluators))
        b = self.with_id(msg.id) \
                .with_task(rv.TaskConfig.from_proto(msg.task)) \
                .with_backend(rv.BackendConfig.from_proto(msg.backend)) \
                .with_dataset(rv.DatasetConfig.from_proto(msg.dataset)) \
                .with_analyzers(analyzers) \
                .with_evaluators(evaluators) \
                .with_root_uri(msg.root_uri) \
                .with_analyze_uri(msg.analyze_uri) \
                .with_chip_uri(msg.chip_uri) \
                .with_train_uri(msg.train_uri) \
                .with_predict_uri(msg.predict_uri) \
                .with_eval_uri(msg.eval_uri) \
                .with_bundle_uri(msg.bundle_uri)

        if msg.custom_config:
            b = b.with_custom_config(json.loads(msg.custom_config['config']))

        return b

    def _copy(self):
        """Create a copy; avoid using deepcopy on the dataset
        as it can have performance implicitions.
        """
        e = ExperimentConfigBuilder()
        e.config['id'] = self.config.get('id')
        e.config['task'] = deepcopy(self.config.get('task'))
        e.config['backend'] = deepcopy(self.config.get('backend'))
        e.config['dataset'] = self.config.get('dataset')
        e.config['analyzers'] = self.config.get('analyzers')
        e.config['evaluators'] = self.config.get('evaluators')
        e.config['root_uri'] = self.config.get('root_uri')
        e.config['analyze_uri'] = self.config.get('analyze_uri')
        e.config['chip_uri'] = self.config.get('chip_uri')
        e.config['train_uri'] = self.config.get('train_uri')
        e.config['predict_uri'] = self.config.get('predict_uri')
        e.config['eval_uri'] = self.config.get('eval_uri')
        e.config['bundle_uri'] = self.config.get('bundle_uri')
        e.config['custom_config'] = self.config.get('custom_config')
        e.analyze_key = self.analyze_key
        e.chip_key = self.chip_key
        e.train_key = self.train_key
        e.predict_key = self.predict_key
        e.eval_key = self.eval_key
        e.bundle_key = self.bundle_key

        return e

    def with_id(self, id):
        """Sets an id for the experiment."""
        b = self._copy()
        b.config['id'] = id
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = self._copy()
        b.config['task'] = task
        return b

    def with_backend(self, backend):
        """Specifies the backend to be used, e.g. rv.TF_DEEPLAB."""
        b = self._copy()
        b.config['backend'] = backend
        return b

    def with_dataset(self, dataset):
        """Specifies the dataset to be used."""
        b = self._copy()
        b.config['dataset'] = dataset
        return b

    def with_analyzers(self, analyzers):
        """Add analyzers to be used in the analysis stage."""
        b = self._copy()
        b.config['analyzers'] = analyzers
        return b

    def with_analyzer(self, analyzer):
        """Add an analyzer to be used in the analysis stage."""
        return self.with_analyzers([analyzer])

    def with_stats_analyzer(self):
        """Add a stats analyzer to be used in the analysis stage."""
        a = rv.AnalyzerConfig.builder(rv.STATS_ANALYZER).build()
        return self.with_analyzer(a)

    def with_evaluators(self, evaluators):
        """Sets the evaluators to use for the evaluation stage."""
        b = self._copy()
        b.config['evaluators'] = evaluators
        return b

    def with_evaluator(self, evaluator):
        """Sets the evaluator to use for the evaluation stage."""
        return self.with_evaluators([evaluator])

    def with_root_uri(self, uri):
        """Sets the root directory where all output will be stored unless
        subsequently overridden.

        """
        b = self._copy()
        b.config['root_uri'] = uri
        return b

    def with_analyze_uri(self, uri):
        """Sets the location where the results of the analysis stage will be
           stored.

        """
        b = self._copy()
        b.config['analyze_uri'] = uri
        return b

    def with_chip_uri(self, uri):
        """Sets the location where the results of the "chip" stage will be
           stored.

        """
        b = self._copy()
        b.config['chip_uri'] = uri
        return b

    def with_train_uri(self, uri):
        """Sets the location where the results of the training stage will be
           stored.

        """
        b = self._copy()
        b.config['train_uri'] = uri
        return b

    def with_predict_uri(self, uri):
        """Sets the location where the results of the prediction stage will be
           stored.

        """
        b = self._copy()
        b.config['predict_uri'] = uri
        return b

    def with_eval_uri(self, uri):
        """Sets the location where the results of the evaluation stage will be
           stored.

        """
        b = self._copy()
        b.config['eval_uri'] = uri
        return b

    def with_bundle_uri(self, uri):
        """Sets the location where the results of the bundling stage will be
           stored.

        """
        b = self._copy()
        b.config['bundle_uri'] = uri
        return b

    def clear_command_uris(self):
        """Clears existing command URIs and keys.
        Useful for re-using experiment configs for new builders.
        """
        b = self._copy()
        b = b.with_analyze_key(None) \
             .with_analyze_uri(None) \
             .with_chip_key(None) \
             .with_chip_uri(None) \
             .with_train_key(None) \
             .with_train_uri(None) \
             .with_predict_key(None) \
             .with_predict_uri(None) \
             .with_eval_key(None) \
             .with_eval_uri(None) \
             .with_bundle_key(None) \
             .with_bundle_uri(None) \

        return b

    def with_custom_config(self, config):
        """Sets custom configuration for this experiment.
        This can be used by plugins such as custom commands."""
        b = self._copy()
        b.config['custom_config'] = config
        return b

    def with_analyze_key(self, key):
        """Sets the key associated with the analysis stage."""
        b = self._copy()
        b.analyze_key = key
        return b

    def with_chip_key(self, key):
        """Sets the key associated with the "chip" stage."""
        b = self._copy()
        b.chip_key = key
        return b

    def with_train_key(self, key):
        """Sets the key associated with the training stage."""
        b = self._copy()
        b.train_key = key
        return b

    def with_predict_key(self, key):
        """Sets the key associated with the prediction stage."""
        b = self._copy()
        b.predict_key = key
        return b

    def with_eval_key(self, key):
        """Sets the key associated with the evaluation stage."""
        b = self._copy()
        b.eval_key = key
        return b

    def with_bundle_key(self, key):
        """Sets the key associated with the bundling stage."""
        b = self._copy()
        b.bundle_key = key
        return b



# flake8: noqa

from os import path as op
import os
import json
import io
import re
from setuptools import (setup, find_packages)
from imp import load_source

__version__ = load_source('rastervision.version',
                          'rastervision/version.py').__version__

here = op.abspath(op.dirname(__file__))

# get the dependencies and installs
with io.open(op.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

# The RTD build environment fails with the reqs in bad_reqs.
if 'READTHEDOCS' in os.environ:
    bad_reqs = ['pyproj', 'h5py']
    all_reqs = list(filter(lambda r: r.split('==')[0] not in bad_reqs, all_reqs))

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [
    x.strip().replace('git+', '') for x in all_reqs if 'git+' not in x
]

def replace_images(readme):
    """Replaces image links in the README with static links to
    the GitHub release branch."""
    release_branch = '.'.join(__version__.split('.')[:2])
    r = r'\((docs/)(.*\.png)'
    rep = (r'(https://raw.githubusercontent.com/azavea/raster-vision/'
           '{}/docs/\g<2>'.format(release_branch))

    return re.sub(r, rep, readme)

# Dependencies for extras, which pertain to installing specific backends.
with io.open(op.join(here, 'extras_requirements.json'), encoding='utf-8') as f:
    extras_require = json.loads(f.read())

# Uncomment this line if we are using a commit of mask-to-polygons
# (as opposed to released version) to avoid error.
# del extras_require['feature-extraction']

setup(
    name='rastervision',
    version=__version__,
    description='An open source framework for deep learning '
    'on satellite and aerial imagery',
    long_description=replace_images(open('README.md').read()),
    long_description_content_type='text/markdown',
    url='https://github.com/azavea/raster-vision',
    author='Azavea',
    author_email='info@azavea.com',
    license='Apache License 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    keywords=
    'raster deep-learning ml computer-vision earth-observation geospatial geospatial-processing',
    packages=find_packages(exclude=['integration_tests*', 'tests*']),
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    dependency_links=dependency_links,
    entry_points='''
        [console_scripts]
        rastervision=rastervision.cli.main:main
    ''',
)



import os
import zipfile

import rastervision as rv
from rastervision.utils.files import (download_if_needed, make_dir,
                                      load_json_config, save_json_config)
from rastervision.protos.command_pb2 import CommandConfig as CommandConfigMsg
from rastervision.data.raster_source import ChannelOrderError


class Predictor():
    """Class for making predictions based off of a prediction package."""

    def __init__(self,
                 prediction_package_uri,
                 tmp_dir,
                 update_stats=False,
                 channel_order=None):
        """Creates a new Predictor.

        Args:
            prediction_package_uri: The URI of the prediction package to use.
                Can be any type of URI that Raster Vision can read.
            tmp_dir: Temporary directory in which to store files that are used
                by the Predictor. This directory is not cleaned up by this
                class.
            update_stats: Option indicating if any Analyzers should be run on
                the image to be predicted on. Otherwise, the Predictor will use
                the output of Analyzers that are bundled with the predict
                package. This is useful, for instance, if you are predicting
                against imagery that needs to be normalized with a
                StatsAnalyzer, and the color profile of the new imagery is
                significantly different then the imagery the model was trained
                on.
            channel_order: Option for a new channel order to use for the
                imagery being predicted against. If not present, the
                channel_order from the original configuration in the predict
                package will be used.
        """
        self.tmp_dir = tmp_dir
        self.update_stats = update_stats
        self.model_loaded = False

        package_zip_path = download_if_needed(prediction_package_uri, tmp_dir)
        package_dir = os.path.join(tmp_dir, 'package')
        make_dir(package_dir)
        with zipfile.ZipFile(package_zip_path, 'r') as package_zip:
            package_zip.extractall(path=package_dir)

        # Read bundle command config
        bundle_config_path = os.path.join(package_dir, 'bundle_config.json')
        msg = load_json_config(bundle_config_path, CommandConfigMsg())
        bundle_config = msg.bundle_config

        self.task_config = rv.TaskConfig.from_proto(bundle_config.task) \
                                        .load_bundle_files(package_dir)

        self.backend_config = rv.BackendConfig.from_proto(bundle_config.backend) \
                                              .load_bundle_files(package_dir)

        scene_config = rv.SceneConfig.from_proto(bundle_config.scene)
        scene_builder = scene_config.load_bundle_files(package_dir) \
                                    .to_builder() \
                                    .clear_label_source() \
                                    .clear_aois() \
                                    .with_id('PREDICTOR')

        # If the scene does not have a label store, generate a default one.
        if not scene_config.label_store:
            scene_builder = scene_builder.with_task(self.task_config) \
                                         .with_label_store()

        if channel_order:
            raster_source = scene_builder.config['raster_source'] \
                                         .to_builder() \
                                         .with_channel_order(channel_order) \
                                         .build()
            scene_builder = scene_builder.with_raster_source(raster_source)

        self.scene_config = scene_builder.build()

        self.analyzer_configs = []
        if update_stats:
            for analyzer in bundle_config.analyzers:
                a = rv.AnalyzerConfig.from_proto(analyzer) \
                                     .load_bundle_files(package_dir)
                self.analyzer_configs.append(a)

        self.bundle_config = rv.command.CommandConfig \
                                       .from_proto(msg) \
                                       .to_builder() \
                                       .with_task(self.task_config) \
                                       .with_backend(self.backend_config) \
                                       .with_scene(self.scene_config) \
                                       .with_analyzers(self.analyzer_configs) \
                                       .build()

    def load_model(self):
        """Load the model for this Predictor.

        This is useful if you are going to make multiple predictions with the
        model, and want it to be fast on the first prediction.

        Note: This is called implicitly on the first call of 'predict' if it
        hasn't been called already.
        """
        self.backend = self.backend_config.create_backend(self.task_config)
        self.backend.load_model(self.tmp_dir)
        self.task = self.task_config.create_task(self.backend)
        self.analyzers = []
        for analyzer_config in self.analyzer_configs:
            self.analyzers.append(analyzer_config.create_analyzer())
        self.model_loaded = True

    def predict(self, image_uri, label_uri=None, config_uri=None):
        """Generate predictions for the given image.

        Args:
            image_uri: URI of the image to make predictions against.
                This can be any type of URI readable by Raster Vision
                FileSystems.
            label_uri: Optional URI to save labels off into.
            config_uri: Optional URI in which to save the bundle_config,
                which can be useful to client applications for understanding
                how to interpret the labels.

            Returns:
                rastervision.data.labels.Labels containing the predicted labels.
        """
        if not self.model_loaded:
            self.load_model()
        scene_config = self.scene_config.for_prediction(image_uri, label_uri) \
                                        .create_local(self.tmp_dir)

        try:
            scene = scene_config.create_scene(self.task_config, self.tmp_dir)
            # If we are analyzing per scene, run analyzers
            # Analyzers should overwrite files in the tmp_dir
            if self.update_stats:
                for analyzer in self.analyzers:
                    analyzer.process([scene], self.tmp_dir)

                # Reload scene to refresh any new analyzer config
                scene = scene_config.create_scene(self.task_config,
                                                  self.tmp_dir)
        except ChannelOrderError:
            raise ValueError(
                'The predict package is using a channel_order '
                'with channels unavailable in the imagery.\nTo set a new '
                'channel_order that only uses channels available in the '
                'imagery, use the --channel-order option.')

        with scene.activate():
            labels = self.task.predict_scene(scene, self.tmp_dir)
            if label_uri:
                scene.prediction_label_store.save(labels)

        if config_uri:
            msg = self.bundle_config.to_builder() \
                                    .with_scene(scene_config) \
                                    .build() \
                                    .to_proto()
            save_json_config(msg, config_uri)

        return labels



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/dataset.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos import scene_pb2 as rastervision_dot_protos_dot_scene__pb2
from rastervision.protos import augmentor_pb2 as rastervision_dot_protos_dot_augmentor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/dataset.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n!rastervision/protos/dataset.proto\x12\trv.protos\x1a\x1frastervision/protos/scene.proto\x1a#rastervision/protos/augmentor.proto\"\xcd\x01\n\rDatasetConfig\x12,\n\x0ctrain_scenes\x18\x03 \x03(\x0b\x32\x16.rv.protos.SceneConfig\x12\x31\n\x11validation_scenes\x18\x04 \x03(\x0b\x32\x16.rv.protos.SceneConfig\x12+\n\x0btest_scenes\x18\x05 \x03(\x0b\x32\x16.rv.protos.SceneConfig\x12.\n\naugmentors\x18\x06 \x03(\x0b\x32\x1a.rv.protos.AugmentorConfig')
  ,
  dependencies=[rastervision_dot_protos_dot_scene__pb2.DESCRIPTOR,rastervision_dot_protos_dot_augmentor__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_DATASETCONFIG = _descriptor.Descriptor(
  name='DatasetConfig',
  full_name='rv.protos.DatasetConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='train_scenes', full_name='rv.protos.DatasetConfig.train_scenes', index=0,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validation_scenes', full_name='rv.protos.DatasetConfig.validation_scenes', index=1,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_scenes', full_name='rv.protos.DatasetConfig.test_scenes', index=2,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='augmentors', full_name='rv.protos.DatasetConfig.augmentors', index=3,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=119,
  serialized_end=324,
)

_DATASETCONFIG.fields_by_name['train_scenes'].message_type = rastervision_dot_protos_dot_scene__pb2._SCENECONFIG
_DATASETCONFIG.fields_by_name['validation_scenes'].message_type = rastervision_dot_protos_dot_scene__pb2._SCENECONFIG
_DATASETCONFIG.fields_by_name['test_scenes'].message_type = rastervision_dot_protos_dot_scene__pb2._SCENECONFIG
_DATASETCONFIG.fields_by_name['augmentors'].message_type = rastervision_dot_protos_dot_augmentor__pb2._AUGMENTORCONFIG
DESCRIPTOR.message_types_by_name['DatasetConfig'] = _DATASETCONFIG

DatasetConfig = _reflection.GeneratedProtocolMessageType('DatasetConfig', (_message.Message,), dict(
  DESCRIPTOR = _DATASETCONFIG,
  __module__ = 'rastervision.protos.dataset_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.DatasetConfig)
  ))
_sym_db.RegisterMessage(DatasetConfig)


# @@protoc_insertion_point(module_scope)



from rastervision.data.label import ChipClassificationLabels
from rastervision.data.label_store import LabelStore
from rastervision.data.label_store.utils import boxes_to_geojson
from rastervision.data.label_source import ChipClassificationLabelSource
from rastervision.data.vector_source import GeoJSONVectorSource
from rastervision.utils.files import json_to_file


class ChipClassificationGeoJSONStore(LabelStore):
    """A GeoJSON file with classification labels in it.
    """

    def __init__(self, uri, crs_transformer, class_map):
        """Construct ClassificationLabelStore backed by a GeoJSON file.

        Args:
            uri: uri of GeoJSON file containing labels
            crs_transformer: CRSTransformer to convert from map coords in label
                in GeoJSON file to pixel coords.
            class_map: ClassMap used to infer class_ids from class_name
                (or label) field
        """
        self.uri = uri
        self.crs_transformer = crs_transformer
        self.class_map = class_map

    def save(self, labels):
        """Save labels to URI if writable.

        Note that if the grid is inferred from polygons, only the grid will be
        written, not the original polygons.
        """
        boxes = labels.get_cells()
        class_ids = labels.get_class_ids()
        scores = list(labels.get_scores())
        geojson = boxes_to_geojson(
            boxes,
            class_ids,
            self.crs_transformer,
            self.class_map,
            scores=scores)
        json_to_file(geojson, self.uri)

    def get_labels(self):
        vector_source = GeoJSONVectorSource(self.uri, self.crs_transformer)
        source = ChipClassificationLabelSource(
            vector_source, self.crs_transformer, self.class_map)
        return source.get_labels()

    def empty_labels(self):
        return ChipClassificationLabels()



import json
import logging
import copy
from subprocess import check_output
import os
from functools import lru_cache

from supermercado.burntiles import burn
from shapely.geometry import shape, mapping
from shapely.ops import cascaded_union

from rastervision.data.vector_source.vector_source import VectorSource
from rastervision.utils.files import get_cached_file
from rastervision.rv_config import RVConfig

log = logging.getLogger(__name__)


def merge_geojson(geojson, id_field):
    """Merge features that share an id.

    Args:
        geojson: (dict) GeoJSON with features that have an id property used to merge
            features that are split across multiple tiles.
        id_field: (str) name of field in feature['properties'] that contains the
            feature's unique id. Used for merging features that are split across
            tile boundaries.
    """
    id_to_features = {}
    for f in geojson['features']:
        id = f['properties'][id_field]
        id_features = id_to_features.get(id, [])
        id_features.append(f)
        id_to_features[id] = id_features

    proc_features = []
    for id, id_features in id_to_features.items():
        id_geoms = []
        for f in id_features:
            g = shape(f['geometry'])
            # Self-intersection trick.
            if f['geometry']['type'] in ['Polygon', 'MultiPolygon']:
                g = g.buffer(0)
            id_geoms.append(g)

        union_geom = cascaded_union(id_geoms)
        union_feature = copy.deepcopy(id_features[0])
        union_feature['geometry'] = mapping(union_geom)
        proc_features.append(union_feature)

    return {'type': 'FeatureCollection', 'features': proc_features}


@lru_cache(maxsize=32)
def get_tile_features(tile_uri, z, x, y):
    """Get GeoJSON features for a specific tile using in-memory caching."""
    cache_dir = os.path.join(RVConfig.get_tmp_dir_root(), 'vector-tiles')
    tile_path = get_cached_file(cache_dir, tile_uri)
    cmd = ['tippecanoe-decode', '-f', '-c', tile_path, str(z), str(x), str(y)]
    log.debug('Running command: {}'.format(' '.join(cmd)))
    tile_geojson_str = check_output(cmd).decode('utf-8')
    tile_features = [
        json.loads(ts) for ts in tile_geojson_str.split('\n')
        if ts.startswith('{')
    ]

    return tile_features


def vector_tile_to_geojson(uri, zoom, map_extent):
    """Get GeoJSON features that overlap with an extent from a vector tile endpoint."""
    log.info('Downloading and converting vector tiles to GeoJSON...')

    # Figure out which tiles cover the extent.
    extent_polys = [{
        'type': 'Feature',
        'properties': {},
        'geometry': {
            'type': 'Polygon',
            'coordinates': [map_extent.geojson_coordinates()]
        }
    }]
    xyzs = burn(extent_polys, zoom)

    # Retrieve tile features.
    features = []
    for xyz in xyzs:
        x, y, z = xyz
        # If this isn't a zxy schema, this is a no-op.
        tile_uri = uri.format(x=x, y=y, z=z)
        tile_features = get_tile_features(tile_uri, z, x, y)
        features.extend(tile_features)

    # Crop features to extent
    extent_geom = map_extent.to_shapely()
    cropped_features = []
    for f in features:
        geom = shape(f['geometry'])
        if f['geometry']['type'].lower() in ['polygon', 'multipolygon']:
            geom = geom.buffer(0)
        geom = geom.intersection(extent_geom)
        if not geom.is_empty:
            f = dict(f)
            f['geometry'] = mapping(geom)
            cropped_features.append(f)

    return {'type': 'FeatureCollection', 'features': cropped_features}


class VectorTileVectorSource(VectorSource):
    def __init__(self,
                 uri,
                 zoom,
                 id_field,
                 crs_transformer,
                 extent,
                 line_bufs=None,
                 point_bufs=None,
                 class_inf_opts=None):
        """Constructor.

        Args:
            uri: (str) URI of vector tile endpoint. Should either contain {z}/{x}/{y} or
                point to .mbtiles file.
            zoom: (int) valid zoom level to use when fetching tiles from endpoint
            id_field: (str) name of field in feature['properties'] that contains the
                feature's unique id. Used for merging features that are split across
                tile boundaries.
            crs_transformer: (CRSTransformer)
            extent: (Box) extent of scene which determines which features to return
            class_inf_opts: (ClassInferenceOptions)
        """
        self.uri = uri
        self.zoom = zoom
        self.id_field = id_field
        self.extent = extent
        super().__init__(crs_transformer, line_bufs, point_bufs,
                         class_inf_opts)

    def _get_geojson(self):
        # This attempts to do things in an efficient order. First, we extract raw
        # GeoJSON from the vector tiles covering the extent. This uses caching, so that
        # we never have to process the same vector tile twice (even across scenes). Then,
        # we infer class ids, which drops any irrelevant features which should speed up
        # the next phase which merges features that are split across tiles.
        map_extent = self.extent.reproject(
            lambda point: self.crs_transformer.pixel_to_map(point))
        log.debug(
            'Reading and converting vector tiles to GeoJSON for extent...')
        geojson = vector_tile_to_geojson(self.uri, self.zoom, map_extent)
        log.debug('Inferring class_ids...')
        geojson = self.class_inference.transform_geojson(geojson)
        log.debug('Merging GeoJSON features...')
        geojson = merge_geojson(geojson, self.id_field)
        return geojson



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/backend.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/backend.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n!rastervision/protos/backend.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\"\xab\x0c\n\rBackendConfig\x12\x14\n\x0c\x62\x61\x63kend_type\x18\x01 \x02(\t\x12\x1c\n\x14pretrained_model_uri\x18\x03 \x01(\t\x12V\n\x1atf_object_detection_config\x18\x04 \x01(\x0b\x32\x30.rv.protos.BackendConfig.TFObjectDetectionConfigH\x00\x12Y\n\x1bkeras_classification_config\x18\x05 \x01(\x0b\x32\x32.rv.protos.BackendConfig.KerasClassificationConfigH\x00\x12\x45\n\x11tf_deeplab_config\x18\x07 \x01(\x0b\x32(.rv.protos.BackendConfig.TFDeeplabConfigH\x00\x12\x30\n\rcustom_config\x18\x06 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x1a\xb6\x03\n\x17TFObjectDetectionConfig\x12\x1a\n\rsync_interval\x18\x01 \x01(\x05:\x03\x36\x30\x30\x12\x1b\n\rdo_monitoring\x18\x02 \x01(\x08:\x04true\x12\x1c\n\rreplace_model\x18\x03 \x01(\x08:\x05\x66\x61lse\x12\x44\n\rmodel_main_py\x18\x04 \x01(\t:-/opt/tf-models/object_detection/model_main.py\x12L\n\texport_py\x18\x05 \x01(\t:9/opt/tf-models/object_detection/export_inference_graph.py\x12\x19\n\x11training_data_uri\x18\x06 \x01(\t\x12\x1b\n\x13training_output_uri\x18\x07 \x01(\t\x12\x11\n\tmodel_uri\x18\x08 \x01(\t\x12!\n\x19\x66ine_tune_checkpoint_name\x18\t \x01(\t\x12\x14\n\x05\x64\x65\x62ug\x18\n \x01(\x08:\x05\x66\x61lse\x12,\n\x0btfod_config\x18\x0b \x02(\x0b\x32\x17.google.protobuf.Struct\x1a\xff\x01\n\x19KerasClassificationConfig\x12\x1a\n\rsync_interval\x18\x01 \x01(\x05:\x03\x36\x30\x30\x12\x1b\n\rdo_monitoring\x18\x02 \x01(\x08:\x04true\x12\x1c\n\rreplace_model\x18\x03 \x01(\x08:\x05\x66\x61lse\x12\x19\n\x11training_data_uri\x18\x04 \x01(\t\x12\x1b\n\x13training_output_uri\x18\x05 \x01(\t\x12\x11\n\tmodel_uri\x18\x06 \x01(\t\x12\x14\n\x05\x64\x65\x62ug\x18\x07 \x01(\x08:\x05\x66\x61lse\x12*\n\tkc_config\x18\x08 \x02(\x0b\x32\x17.google.protobuf.Struct\x1a\xec\x03\n\x0fTFDeeplabConfig\x12\x31\n\x08train_py\x18\x01 \x01(\t:\x1f/opt/tf-models/deeplab/train.py\x12/\n\x07\x65val_py\x18\x0e \x01(\t:\x1e/opt/tf-models/deeplab/eval.py\x12\x39\n\texport_py\x18\x02 \x01(\t:&/opt/tf-models/deeplab/export_model.py\x12\x19\n\x11train_restart_dir\x18\x03 \x01(\t\x12\x1a\n\rsync_interval\x18\x04 \x01(\x05:\x03\x36\x30\x30\x12\x1b\n\rdo_monitoring\x18\x05 \x01(\x08:\x04true\x12\x16\n\x07\x64o_eval\x18\r \x01(\x08:\x05\x66\x61lse\x12\x1c\n\rreplace_model\x18\x06 \x01(\x08:\x05\x66\x61lse\x12\x19\n\x11training_data_uri\x18\x07 \x01(\t\x12\x1b\n\x13training_output_uri\x18\x08 \x01(\t\x12\x11\n\tmodel_uri\x18\t \x01(\t\x12!\n\x19\x66ine_tune_checkpoint_name\x18\n \x01(\t\x12,\n\x0btfdl_config\x18\x0b \x02(\x0b\x32\x17.google.protobuf.Struct\x12\x14\n\x05\x64\x65\x62ug\x18\x0c \x01(\x08:\x05\x66\x61lseB\x10\n\x0e\x62\x61\x63kend_config')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BACKENDCONFIG_TFOBJECTDETECTIONCONFIG = _descriptor.Descriptor(
  name='TFObjectDetectionConfig',
  full_name='rv.protos.BackendConfig.TFObjectDetectionConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sync_interval', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.sync_interval', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=600,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='do_monitoring', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.do_monitoring', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='replace_model', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.replace_model', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model_main_py', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.model_main_py', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("/opt/tf-models/object_detection/model_main.py").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='export_py', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.export_py', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("/opt/tf-models/object_detection/export_inference_graph.py").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_data_uri', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.training_data_uri', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_output_uri', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.training_output_uri', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model_uri', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.model_uri', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fine_tune_checkpoint_name', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.fine_tune_checkpoint_name', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.debug', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tfod_config', full_name='rv.protos.BackendConfig.TFObjectDetectionConfig.tfod_config', index=10,
      number=11, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=449,
  serialized_end=887,
)

_BACKENDCONFIG_KERASCLASSIFICATIONCONFIG = _descriptor.Descriptor(
  name='KerasClassificationConfig',
  full_name='rv.protos.BackendConfig.KerasClassificationConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sync_interval', full_name='rv.protos.BackendConfig.KerasClassificationConfig.sync_interval', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=600,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='do_monitoring', full_name='rv.protos.BackendConfig.KerasClassificationConfig.do_monitoring', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='replace_model', full_name='rv.protos.BackendConfig.KerasClassificationConfig.replace_model', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_data_uri', full_name='rv.protos.BackendConfig.KerasClassificationConfig.training_data_uri', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_output_uri', full_name='rv.protos.BackendConfig.KerasClassificationConfig.training_output_uri', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model_uri', full_name='rv.protos.BackendConfig.KerasClassificationConfig.model_uri', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug', full_name='rv.protos.BackendConfig.KerasClassificationConfig.debug', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kc_config', full_name='rv.protos.BackendConfig.KerasClassificationConfig.kc_config', index=7,
      number=8, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=890,
  serialized_end=1145,
)

_BACKENDCONFIG_TFDEEPLABCONFIG = _descriptor.Descriptor(
  name='TFDeeplabConfig',
  full_name='rv.protos.BackendConfig.TFDeeplabConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='train_py', full_name='rv.protos.BackendConfig.TFDeeplabConfig.train_py', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("/opt/tf-models/deeplab/train.py").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_py', full_name='rv.protos.BackendConfig.TFDeeplabConfig.eval_py', index=1,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("/opt/tf-models/deeplab/eval.py").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='export_py', full_name='rv.protos.BackendConfig.TFDeeplabConfig.export_py', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("/opt/tf-models/deeplab/export_model.py").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_restart_dir', full_name='rv.protos.BackendConfig.TFDeeplabConfig.train_restart_dir', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sync_interval', full_name='rv.protos.BackendConfig.TFDeeplabConfig.sync_interval', index=4,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=600,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='do_monitoring', full_name='rv.protos.BackendConfig.TFDeeplabConfig.do_monitoring', index=5,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='do_eval', full_name='rv.protos.BackendConfig.TFDeeplabConfig.do_eval', index=6,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='replace_model', full_name='rv.protos.BackendConfig.TFDeeplabConfig.replace_model', index=7,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_data_uri', full_name='rv.protos.BackendConfig.TFDeeplabConfig.training_data_uri', index=8,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_output_uri', full_name='rv.protos.BackendConfig.TFDeeplabConfig.training_output_uri', index=9,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model_uri', full_name='rv.protos.BackendConfig.TFDeeplabConfig.model_uri', index=10,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fine_tune_checkpoint_name', full_name='rv.protos.BackendConfig.TFDeeplabConfig.fine_tune_checkpoint_name', index=11,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tfdl_config', full_name='rv.protos.BackendConfig.TFDeeplabConfig.tfdl_config', index=12,
      number=11, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug', full_name='rv.protos.BackendConfig.TFDeeplabConfig.debug', index=13,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1148,
  serialized_end=1640,
)

_BACKENDCONFIG = _descriptor.Descriptor(
  name='BackendConfig',
  full_name='rv.protos.BackendConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='backend_type', full_name='rv.protos.BackendConfig.backend_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pretrained_model_uri', full_name='rv.protos.BackendConfig.pretrained_model_uri', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tf_object_detection_config', full_name='rv.protos.BackendConfig.tf_object_detection_config', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='keras_classification_config', full_name='rv.protos.BackendConfig.keras_classification_config', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tf_deeplab_config', full_name='rv.protos.BackendConfig.tf_deeplab_config', index=4,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.BackendConfig.custom_config', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_BACKENDCONFIG_TFOBJECTDETECTIONCONFIG, _BACKENDCONFIG_KERASCLASSIFICATIONCONFIG, _BACKENDCONFIG_TFDEEPLABCONFIG, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='backend_config', full_name='rv.protos.BackendConfig.backend_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=79,
  serialized_end=1658,
)

_BACKENDCONFIG_TFOBJECTDETECTIONCONFIG.fields_by_name['tfod_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_BACKENDCONFIG_TFOBJECTDETECTIONCONFIG.containing_type = _BACKENDCONFIG
_BACKENDCONFIG_KERASCLASSIFICATIONCONFIG.fields_by_name['kc_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_BACKENDCONFIG_KERASCLASSIFICATIONCONFIG.containing_type = _BACKENDCONFIG
_BACKENDCONFIG_TFDEEPLABCONFIG.fields_by_name['tfdl_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_BACKENDCONFIG_TFDEEPLABCONFIG.containing_type = _BACKENDCONFIG
_BACKENDCONFIG.fields_by_name['tf_object_detection_config'].message_type = _BACKENDCONFIG_TFOBJECTDETECTIONCONFIG
_BACKENDCONFIG.fields_by_name['keras_classification_config'].message_type = _BACKENDCONFIG_KERASCLASSIFICATIONCONFIG
_BACKENDCONFIG.fields_by_name['tf_deeplab_config'].message_type = _BACKENDCONFIG_TFDEEPLABCONFIG
_BACKENDCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_BACKENDCONFIG.oneofs_by_name['backend_config'].fields.append(
  _BACKENDCONFIG.fields_by_name['tf_object_detection_config'])
_BACKENDCONFIG.fields_by_name['tf_object_detection_config'].containing_oneof = _BACKENDCONFIG.oneofs_by_name['backend_config']
_BACKENDCONFIG.oneofs_by_name['backend_config'].fields.append(
  _BACKENDCONFIG.fields_by_name['keras_classification_config'])
_BACKENDCONFIG.fields_by_name['keras_classification_config'].containing_oneof = _BACKENDCONFIG.oneofs_by_name['backend_config']
_BACKENDCONFIG.oneofs_by_name['backend_config'].fields.append(
  _BACKENDCONFIG.fields_by_name['tf_deeplab_config'])
_BACKENDCONFIG.fields_by_name['tf_deeplab_config'].containing_oneof = _BACKENDCONFIG.oneofs_by_name['backend_config']
_BACKENDCONFIG.oneofs_by_name['backend_config'].fields.append(
  _BACKENDCONFIG.fields_by_name['custom_config'])
_BACKENDCONFIG.fields_by_name['custom_config'].containing_oneof = _BACKENDCONFIG.oneofs_by_name['backend_config']
DESCRIPTOR.message_types_by_name['BackendConfig'] = _BACKENDCONFIG

BackendConfig = _reflection.GeneratedProtocolMessageType('BackendConfig', (_message.Message,), dict(

  TFObjectDetectionConfig = _reflection.GeneratedProtocolMessageType('TFObjectDetectionConfig', (_message.Message,), dict(
    DESCRIPTOR = _BACKENDCONFIG_TFOBJECTDETECTIONCONFIG,
    __module__ = 'rastervision.protos.backend_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.BackendConfig.TFObjectDetectionConfig)
    ))
  ,

  KerasClassificationConfig = _reflection.GeneratedProtocolMessageType('KerasClassificationConfig', (_message.Message,), dict(
    DESCRIPTOR = _BACKENDCONFIG_KERASCLASSIFICATIONCONFIG,
    __module__ = 'rastervision.protos.backend_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.BackendConfig.KerasClassificationConfig)
    ))
  ,

  TFDeeplabConfig = _reflection.GeneratedProtocolMessageType('TFDeeplabConfig', (_message.Message,), dict(
    DESCRIPTOR = _BACKENDCONFIG_TFDEEPLABCONFIG,
    __module__ = 'rastervision.protos.backend_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.BackendConfig.TFDeeplabConfig)
    ))
  ,
  DESCRIPTOR = _BACKENDCONFIG,
  __module__ = 'rastervision.protos.backend_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.BackendConfig)
  ))
_sym_db.RegisterMessage(BackendConfig)
_sym_db.RegisterMessage(BackendConfig.TFObjectDetectionConfig)
_sym_db.RegisterMessage(BackendConfig.KerasClassificationConfig)
_sym_db.RegisterMessage(BackendConfig.TFDeeplabConfig)


# @@protoc_insertion_point(module_scope)
